;********************************************************************************************
; ATtiny45_FM_Radio: UKW Stereo Empfnger mit dem Modul RDA5807M mit OLED I2C Display 64x128
; und 3 Tasten zur Steuerung Ressourcen: ATtiny45 @8MHz intern
; Fuses Lo=E2, Hi=DF, Pinbelegung: PB0=SDA, PB1=SCL, PB2=Key-up, PB3=KeyDown, PB4=Key-Mode
;
;								 RDA5807-Modul
;								 -------------
;							SDA |  6  +-+  5  |	Antenne
;							SCL |  7  +-+  4  | NC
;							NC  |  8       3  | NF links
;							NC  |  9 _____ 2  | NF rechts
;						 +3.3V  | 10 Quarz 1  | GND
;								 -------------
;
; V1.75, (c) T. Neveling					 - 26/08/2020
; V1.76, Zeitzonen Offset bei der RDS Zeitabfrage bercksichtigt - 23/01/2021
;********************************************************************************************
;
.NOLIST
.include "tn45def.inc"			; Definitionsdatei fr ATtiny45 CPU einbinden
.LIST
;
.undef	XL						; r26/27 werden nicht als Zeiger benutzt.
.undef	XH
.def	HPOS		= r8		; Zeiger fr Radiotext Ausgabe
.def	MIN			= r9		; RDS Uhrzeit
.def	HOUR		= r10
.def	KeyRep1		= r11		; Fr Tasten Warteschleifen
.def	KeyRep2		= r13
.def	ticks32ms	= r14		; Interrupt Zhlvariable (fr Delays)
.def	IntReg		= r15
;
.def	temp1		= r16		; Universalvariablen
.def	temp2		= r17
.def	temp3		= r18
.def	Counter		= r19		; fr Zhlschleifen
.def	oled_x		= r20		; Cursor Position 0..15 (horizontal)
.def	oled_y		= r21		; Cursor Position 0.. 7 (vertikal)
.def	Flags		= r22
	.equ	Keys	= 0			; gesetzt, wenn Taste gedrckt wurde
	.equ	Stereo	= 2			; gesetzt bei Stereo Signal
;
; Steuerbytes fr RDA5807 im sequeziellen Modus (I2C Adresse = 0x10)
.def	CtrlSeek	= r23		; Control & Seek
.def	Volume		= r24		; Lautstrke (0..15)
.def	Chan_Hi		= r25		; einzustellender Channel (10 Bit)
.def	Chan_Lo		= r26
.def	RSSI		= r12		; Receive signal strength indicator
.def	Mode		= r27		; 0=Volume, 1=Mute, 2=Seek (= Register XL)
;
; r28/29 = YL/YH zur OLED Offsetberechnung bei der Zeichenausgabe
; r30/31 = ZL/ZH als Pointer auf Textstrings & OLED Init-Tabelle
;
.dseg							; Datenbereich im SRAM
.org	SRAM_START
;
D01:	.BYTE 1					; x1
D02:	.BYTE 1					; x10	Dezimalwerte vom 16 Bit HEX => Dezimalwandler
D03:	.BYTE 1					; x100	Maximalwert theoretisch 65535 (5-Stellen)
D04:	.BYTE 1					; x1000
D05:	.BYTE 1					; x10000, hier immer = 0
;
R0AHi:	.BYTE 1					; Read Register A (High Byte)
R0ALo:	.BYTE 1					; Read Register A (Low Byte)
R0BHi:	.BYTE 1					; Read Register B (High Byte)
R0BLo:	.BYTE 1					; Read Register B (Low Byte)
BLKAHi:	.BYTE 1					; RDS Block A (High Byte) = Read Register C
BLKALo:	.BYTE 1					; RDS Block A (Low Byte)
BLKBHi:	.BYTE 1					; RDS Block B (High Byte) = Read Register D
BLKBLo:	.BYTE 1					; RDS Block B (Low Byte)
BLKCHi:	.BYTE 1					; RDS Block C (High Byte) = Read Register E
BLKCLo:	.BYTE 1					; RDS Block C (Low Byte)
BLKDHi:	.BYTE 1					; RDS Block D (High Byte) = Read Register F
BLKDLo:	.BYTE 1					; RDS Block D (Low Byte)
;
FLo:	.BYTE 1					; eingestellte Frequenz im 100kHz Raster (Low/High)
FHi:	.BYTE 1

RDSTXT:	.BYTE 82				; RDS Textpuffer (0..79) 16 Blanks + 64 Zeichen = 80
;
.cseg							; Codesegment
;
; Interrupt Vektoren
.org 0 			rjmp start		; Reset Handler, Startadresse 0
.org OVF0Addr	rjmp t0int		; 32,678ms Overflow Interrupt: Tastenabfrage / Delays
;
start:							; Stackpointer per Default auf RAMEND, s. Datenblatt
;
; Ports initialisieren
	ldi temp1, (1 << 2) | (1 << 3) | (1 << 4)	; 0b00011100 = 28d
	out PORTB, temp1			; Pullup Widerstnde an PB2 bis PB4 (Tasten)
	rcall i2c_init
;
; RDA5807 initialisieren: Variable Bytes vorbesetzen
	ldi CtrlSeek, 0b11010000	; Mute off, Output normal, Stereo, Bass Boost
	ldi Volume, 0x03			; kleine Lautstrke einstellen (Range 0 - 15)
;
	clr Flags					; Kein Tastendruck, Stereo Indikator gelscht
	clr Mode					; Start im Modus "Volume"
	mov ticks32ms, temp1		; Delay-Zhler = 28 aus Port Initialisierung
	rcall ClearBuffer			; RDS Textpuffer lschen
;
; OLED Display initialisieren, lschen, Cursor home
	rcall OLED_init
	rcall OLED_Home
	ldi temp1, 'A'				; Firmware Version ausgeben: A(ssembler)X.YZ
	rcall OLED_PutBigChar
	ldi temp1, '1'
	rcall OLED_PutBigChar
	ldi temp1, '.'
	rcall OLED_PutBigChar
	ldi temp1, '7'
	rcall OLED_PutBigChar
	ldi temp1, '6'
	rcall OLED_PutBigChar
;
; Bildaufbau					  ##  Cursor positionieren  ##
	ldi oled_x, 12				; 0 = links ... 15 = rechts
	ldi oled_y, 0				; 0 = oben.......7 = unten, 8 = oben...
	rcall OLED_GotoXY
	ldi ZL, LOW(Text1*2)		; "MHz"
	ldi ZH, HIGH(Text1*2)		; ZH = 0, da Tabellenadresse < 256
	rcall OLED_PutString
	ldi oled_x, 0
	ldi oled_y, 4
	rcall OLED_GotoXY
	ldi ZL, LOW(Text2*2)		; "Mode: "
;	ldi ZH, HIGH(Text2*2)		; bleibt = 0
	rcall OLED_PutString
	ldi ZL, LOW(Mode0*2)		; "Volume"
;	ldi ZH, HIGH(Mode0*2)		; bleibt = 0
	rcall OLED_PutString
;
	ldi oled_y, 5				; oled_x = 0 von oben
	rcall OLED_GotoXY
	ldi ZL, LOW(Text3*2)		; "Vol.:"
;	ldi ZH, HIGH(Text3*2)		; bleibt = 0
	rcall OLED_PutString
;
	ldi oled_y, 6				; oled_x = 0 von oben
	rcall OLED_GotoXY
	ldi ZL, LOW(Text4*2)		; "Sig.:"
;	ldi ZH, HIGH(Text4*2)		; bleibt = 0
	rcall OLED_PutString
	rcall Set_vol_bar			; Lautstrke-Balken
;
; Timer0: 32,768ms Overflow Interrupt zur Tastenenabfrage und Entprellung
	ldi temp1, (1<<CS02) | (1<<CS00)
	out tccr0b, temp1			; Prescaler 1/1024: 1024 * 256 / 8MHz =  32,768ms
	ldi temp1, 1<<TOIE0
	out timsk, temp1			; Timer0 Interrupt freigeben
	sei							; globale Interrupt Freigabe
;
ini_lp:							; RDA5807 Hardware Reset abwarten!
	tst ticks32ms				; 28 x 32,768ms = 917,5ms
	brne ini_lp					; nachfolgend Enable & Startfrequenz bermitteln
;
	cli							; EEPROM Zugriffe ohne Interrupts!
	ldi temp2, Last_FM + 4		; Startfrequenz aus EEPROM lesen: temp2 = Adresse
	rcall EE_Read				; Byte nach temp1
	sts FLo, temp1				; in SRAM
	inc temp2					; zweites Byte
	rcall EE_Read
	sts FHi, temp1				; in SRAM
	rcall fm_tune				; Enable RDA5807 & Startfrequenz einstellen
	lds temp1, FLo				; temp1/2 neu laden (wurden in fm_tune gendert)
	lds temp2, FHi
	rcall BIN2BCD_16_Out		; Frequenzanzeige
	sei							; Interrupts wieder frei
;
main:							; ### Main Schleife gibt RDS Text aus ###
	rcall Delay_125ms			; ca 125ms Pause
	rcall RDS_Text_Out
	inc HPos
	sbrs HPos, 6				; 64 + 15 = 79 = Obergrenze der Zeichenkette
	rjmp main
	clr HPos
	rjmp main
;
Text1:	.db "MHz", 0			; ### Text Strings / Tabellen ###
Text2:	.db "Mode: ", 0,0
Text3:	.db	"Vol.:", 0
Text4:	.db	"Sig.:", 0
Mode0:	.db "Volume", 0,0
Mode1:	.db	"Mute  ", 0,0
Mode2:	.db "Seek  ", 0,0
Blank:	.db	"                ",0,0	; 16 Blanks fr OLED_ClrScr
Initab:	.db	0x8D, 0x14, 0xAF, 0xA1, 0xC0, 0x00	; OLED Init Sequenz
;
;*************************************************************
;* Timer Interrupt zur Tasten Auswertung, Iteration 32,768ms *
;*************************************************************
t0int:
	push temp1
	in intreg, sreg
	push intreg
;
	rcall Get_RDS_Data			; RDS Register auslesen (30,5 mal / Sekunde)
;
	sbrc Flags, Keys
	rjmp KeysNotZero
								; beim letzten Mal war keine Taste gedrckt
	sbis PINB, 2				; Tasten neu abfragen
	rcall Key_up_pressed		; Btn_up wurde gedrckt
	sbis PINB, 3
	rcall Key_dn_pressed		; Btn_dn wurde gedrckt
	sbis PINB, 4
	rcall Key_md_pressed		; Btn_md wurde gedrckt
	rjmp t0intx
;
KeysNotZero:					; Keys waren beim letzten Mal <> 0: jetzt neu prfen
	cbr Flags, 1<<Keys			; Tastenflag zunchst wieder lschen
	sbis PINB, 2
	sbr Flags, 1<<Keys
	sbis PINB, 3				; wenn keine Taste mehr gedrckt ist bleibt Keys = 0
	sbr Flags, 1<<Keys
	sbis PINB, 4
	sbr Flags, 1<<Keys
;
t0intx:
	dec	ticks32ms				; Delay Zhler dekrementieren
	tst ticks32ms
	brne t0xx
	rcall Get_RSSI				; alle 8,35s die Feldstrke anzeigen (256 x 32,768ms)
t0xx:
	pop intreg
	out sreg, intreg
	pop temp1
	reti
;
;*****************
; RDS Funktionen *
;*****************
ClearBuffer:					; RDS Textpuffer mit Blanks initialisieren
	ldi ZL, low(RDSTXT)
	ldi ZH, high(RDSTXT)		; hier immer = 0, da < 256 Byte!
	ldi Counter, 81
	ldi temp1, ' '
BlankLoop:
	st z+, temp1
	dec Counter
	brne BlankLoop
	clr HPos					; Zeiger auf Anfang
	ret
;
RDS_Text_Out:					; ### Radiotext ausgeben ###
	cli							; hier keine Interrupts!
	ldi oled_x, 0
	ldi oled_y, 7				; Anfang von Zeile 7
	rcall OLED_GotoXY
	ldi zl, Low(RDSTXT)
	ldi zh, high(RDSTXT)		; Zeiger auf Start des Textpuffers
	add zl, HPos				; Zeiger auf 1. auszugebendes Zeichen
;
	ldi temp3, 16				; jeweils 16 Zeichen ab HPos
RDS_Text_Out_LP:
	ld temp1, z+
	andi temp1, 0b01111111		; ASCII Code bis 127
	cpi temp1, ' '
	brsh RDS_Text_Out_1
	ldi temp1, ' '				; keine Zeichen < Blank
RDS_Text_Out_1:
	push zl						; ZL/H werden in OLED_PutChar verndert.
	push zh
	rcall OLED_PutChar
	pop zh
	pop zl
	dec temp3					; Zeichenzhler
	brne RDS_Text_Out_LP
	sei							; Interrupts freigeben
	ret
;
ZweiZiffern:					; 1 Byte aus temp3 als 2 Ziffern auf OLED ausgeben
	ldi temp1, -1 +'0'			; 10er Stelle fr ASCII Ausgabe initialisieren
Lx:	inc temp1					; Subtraktionen zhlen
	subi temp3, 10				; sooft 10 abziehen bis Negativ-berlauf entsteht
	brcc Lx
;
	rcall OLED_PutBigChar		; 10er Stelle aus temp1 ausgeben
	subi temp3, -10	-'0'		; letzte Subtraktion rckgngig machen
	mov temp1, temp3			; umspeichern
	rcall OLED_PutBigChar		; 1er Stelle aus temp1 ausgeben
	ret
;
Get_RDS_Data:					; ### RDS Infos auslesen ###
	rcall fm_read				; RDA5807 Lese-Register ins SRAM kopieren
;
	lds temp1, R0BLo
	andi temp1, 0b00000011		; Block B Fehlerbits sollten beide = 0 sein
	brne RDS_y					; Fehler > 0, fertig, keine Ausgabe
;
	lds temp1, BLKBHi			; Gruppe 4A (Zeit) in Block B abfragen
	andi temp1, 0b11110000		; High Nibble im High Byte sollte den Wert 4 haben
	cpi temp1,  0b01000000
	brne RDS_txt				; Wert <> 4, prfen, ob RDS Text aktiv ist
;
;******************
; Zeit decodieren *
;******************
; 6 Minuten Bits extrahieren aus Block D, Bits 6 - 11
	lds temp1, BLKDLo			; BLOCKD holen (Low & High)
	lds temp2, BLKDHi
	andi temp2, 0x0F			; obere 4 Stundenbits lschen
	ldi counter, 6				; Zahl der Schiebeoperationen
min_LP:
	lsr temp2
	ror temp1
	dec counter
	brne min_LP
	mov MIN, temp1				; Minuten sind fertig.
;
; 5 Stunden Bits in Block D, Bits 12 - 15 und Block C Bit 0
	lds temp1, BLKDHi
	swap temp1					; Stunden Bits 0-3 aus BLOCKD rechtsbndig
	andi temp1, 0x0F			; obere Bits lschen
	lds temp2, BLKCLo			; oberes Bit in BLOCKC Register
	bst temp2, 0				; ber das T-Flag umkopieren
	bld temp1, 4
;
; lokalen Offset bercksichtigen (MESZ etc.)
	lds temp2, BLKDLo
	andi temp2, 0x3F			; i_ofs = BlkD & 0x003F
	lsr temp2					; i_ofs = i_ofs / 2
	add temp1, temp2
	mov HOUR, temp1
;
	ldi oled_x, 0				; ### Ergebnis ausgeben ###
	ldi oled_y, 2
	rcall OLED_GotoXY			; Spalte 0, Zeile 4
	mov temp3, HOUR
	rcall ZweiZiffern			; Stunden als 2 Zahlen ausgeben
	ldi temp1, ':'
	rcall OLED_PutBigChar
	mov temp3, MIN				; Minuten als 2 Zahlen ausgeben
	rcall ZweiZiffern
RDS_y:
	ret
;
;***********************
; Radiotext decodieren *
;***********************
RDS_txt:
	cpi temp1,  0b00100000		; BlkBHi: Gruppe 2A, Radio Text ?
	brne RDS_TA					; Block 0A Traffic Announcement prfen
;
; Radio Text
	ldi ZL, low(RDSTXT)			; Beginn des RDS Textpuffers
	ldi ZH, high(RDSTXT)		; hier immer = 0, da < 256 Byte!
	lds temp1, BlkBLo			; Text Segment Code fr die folgenden 4 Zeichen
	andi temp1, 0x0F			; unteres Nibble enthlt den Code
	lsl temp1
	lsl temp1					; x 4, da 4 Zeichen / Segment gesendet werden
	subi temp1, -16				; 16 feste Blanks links fr Scroll von rechts nach links
	add ZL, temp1				; Zeiger auf Speichersegment innerhalb von RDSTXT (64 Bit)
;
	lds temp1, BlkCHi			; 4 Zeichen aus Block C & D dort abspeichern
	st z+, temp1				; speichern + post Inkrement Adresse
	lds temp1, BlkCLo
	st z+, temp1
	lds temp1, BlkDHi
	st z+, temp1
	lds temp1, BlkDLo
	st z, temp1					; speichern ohne post Inkrement Adresse
;
;	ldi oled_x, 15				; Kontrolle A/B Toggle
;	ldi oled_y, 6
;	rcall OLED_GotoXY
;	ldi temp1, '*'
;	sbrc Flags, TextB
;	ldi temp1, ' '
;	rcall oled_putchar
;
RDS_TA:
	cpi temp1, 0				; Durchsagen-Kennung ?
	brne RDS_x
	ldi oled_x, 11
	ldi oled_y, 2
	rcall OLED_GotoXY
	ldi temp1, ' '
	lds temp3, BlkBLo	
	sbrc temp3, 4				; TA Bit (Traffic Announcement)
	ldi temp1, '!'
	rcall OLED_PutBigChar
;
RDS_x:							; Die Ausgabe erfolgt in der Main-Schleife.
	ret
;
Delay_125ms:					; ca. 125ms @8MHz fr Key Repeat & Laufschrift
	ldi temp1, 5				; #1
WL0:clr	KeyRep1					; #1
WL1:clr KeyRep2					; #1
WL2:dec KeyRep2					; #1	innere Schleife: 768 Zyklen
	brne WL2					; #2	= 96s
	dec KeyRep1					; #1	mittlere Schleife: 772x256 Zyklen
	brne WL1					; #2	= 24,7ms
	dec temp1					; #1	uere Schleife
	brne WL0					; #2	5x24,7ms = 123ms
	ret							; #4
;
;**************
;* Taste "up" *
;**************
Key_up_pressed:
	sbr Flags, 1<<Keys			; Tastenflag setzen
	cpi Mode, 0					; Modus ermitteln
	breq Key_up_1				; Volume
;
	cpi Mode, 2
	breq Key_s_up
;
	rjmp Key_up_x				; im Modus 1 "Mute" keine Aktion
;
;## Modus 2: Seek up
Key_s_up:
	ldi CtrlSeek, 0xD3			; Bits 0 & 1 Seek up setzen
	rcall fm_seek_wait			; Seek/Tune Complete Bit abwarten dann Frequenz einlesen
	rjmp Key_up_x
;
Key_up_1:
;## Modus 0: Volume up
	inc Volume					; Lautstrke hoch
	cpi Volume, 16				; max 15
	brne Key_up_11
	dec Volume
Key_up_11:
	rcall Set_Vol_Bar			; Balkenanzeige mit Wert von Volume
	rcall fm_command			; Lautstrke in RDA5807 setzen
	rcall Delay_125ms			; 125ms
	sbis PINB, 2				; Taste hier repetierend
	rjmp Key_up_1
Key_up_x:						; exit Label
	ret
;
;****************
;* Taste "down" *
;****************
Key_dn_pressed:
	sbr Flags, 1<<Keys			; Tastenflag setzen
	cpi Mode, 0					; Modus ermitteln
	breq Key_dn_1				; Volume
;
	cpi Mode, 2
	breq Key_s_dn
;
	rjmp Key_dn_x				; im Modus 1 "Mute" keine Aktion
;
;## Modus 2: Seek down
Key_s_dn:
	ldi CtrlSeek, 0xD1			; Bit 0 Seek down setzen
	rcall fm_seek_wait			; Seek/Tune Complete Bit abwarten dann Frequenz einlesen
	rjmp Key_dn_x
;
Key_dn_1:
;## Modus 0: Volume down
	dec Volume					; Lautstrke ab
	cpi Volume, 255				; min 0
	brne Key_dn_11
	clr Volume
Key_dn_11:
	rcall Set_Vol_Bar			; Balkenanzeige mit Wert von Volume
	rcall fm_command			; Lautstrke in RDA5807 setzen
	rcall Delay_125ms			; 125ms
	sbis PINB, 3				; Taste hier repetierend
	rjmp Key_dn_1
Key_dn_x:						; exit Label
	ret
;
;****************
;* Taste "Mode" *
;****************
Key_md_pressed:
	sbr Flags, 1<<Keys			; Tastenflag setzen
;
	ldi oled_x, 6				; Spalte 6
	ldi oled_y, 4				; Zeile 4
	rcall OLED_GotoXY
	ldi CtrlSeek, 0b11010000	; Mute OFF, Output normal, Stereo, Bass Boost
;
	inc Mode					; auf 0..2 eingrenzen
	cpi Mode, 3
	brne Mode_ok
	clr Mode
Mode_ok:
	cpi Mode, 0
	brne Key_md1
;
Key_md0:						; Modus 0: "Volume"
	ldi ZL, LOW(Mode0*2)
	ldi ZH, HIGH(Mode0*2)
	rjmp Key_mdx
;
Key_md1:
	cpi Mode, 1
	brne Key_md2
;
	ldi ZL, LOW(Mode1*2)		; Modus 1: "Mute  "
	ldi ZH, HIGH(Mode1*2)
	ldi CtrlSeek, 0b10010000	; Mute ON, Output normal, Stereo, Bass Boost
	rjmp Key_mdx
;
Key_md2:
	ldi ZL, LOW(Mode2*2)		; Modus 2: "Seek  "
	ldi ZH, HIGH(Mode2*2)
;
Key_mdx:						; exit Label
	rcall OLED_PutString		; entsprechenden String ausgeben
	rcall fm_command			; Mute Funktion ein- oder ausschalten
	ret
;
;**************************************
;* Ansteuerung des Lautstrke Balkens *
;**************************************
Set_Vol_Bar:
	ldi oled_x, 6				; Spalte 6
	ldi oled_y, 5				; Zeile 5
	rcall OLED_GotoXY			; Cursor auf Balken Anfang
	mov temp2, Volume			; 4 Bit, Wert 0..15
	lsl temp2
	lsl temp2					; x 4 (15 x 4 = 60 Rasterlinien)
	clr counter					; Rasterlinien 0-60 zhlen
	rcall OLED_start_data		; Datenmodus starten
	ldi temp1, 0x7E				; die kleinste Lautstrke ist >0
	rcall i2c_write				; => links 2 extra Balken einzeichnen
	rcall i2c_write
Vol_LP:
	cp temp2, counter
	brlo v1
	ldi temp1, 0x7E				; Pixelspalte setzen
	rjmp v2
v1:	ldi temp1, 0x42				; Pixelspalte lschen
v2:	rcall i2c_write
	inc counter
	cpi counter, 61
	brne Vol_LP
	ldi temp1, 0x7E				; Pixelspalte setzen
	rcall i2c_write				; Rechteck rechts schlieen
	rcall i2c_stop				; Datenmodus beenden
	ret
;
;***************************************
;* Feldstrke Balken aus RSSI erzeugen *
;***************************************
Get_RSSI:
	ldi oled_x, 6				; Spalte 6
	ldi oled_y, 6				; Zeile 6
	rcall OLED_GotoXY
	clr counter
	mov temp2, RSSI				; Wertebereich 0..63
	rcall OLED_start_data		; Datenmodus starten
RSSI_LP:
	cp temp2, counter
	brlo rs1
	ldi temp1, 0x7E				; Pixelspalte setzen
	rjmp rs2
rs1:ldi temp1, 0x42				; Pixelspalte lschen
rs2:rcall i2c_write
	inc counter
	cpi counter, 63
	brne RSSI_LP
	ldi temp1, 0x7E				; Pixelspalte setzen
	rcall i2c_write				; Rechteck rechts schlieen
	rcall i2c_stop
								; ### Stereo Indikator ###
	ldi oled_x, 13				; Spalte 13
	ldi oled_y, 2				; Zeile 2
	rcall OLED_GotoXY
	ldi temp1, ' '				; mono
	sbrc Flags, Stereo
	ldi temp1, 92				; stereo ("OO" Sonderzeichen)
	rcall OLED_PutBigChar
	ret
;
;***************************************************************
;* Ansteuerung das RDA5807 Radio Moduls im sequentiellem Modus *
;***************************************************************
fm_command:						; Das Beschreiben der 16 Bit Register erfolgt ab R02
	push temp1
	rcall i2c_start				; aufsteigend, immer erst High Byte, dann Low Byte
	ldi temp1, (0x10 << 1) | 0	; I2C Adresse fr sequentiellen Zugriff (schreibend)
	rcall i2c_write
	mov temp1, CtrlSeek			; 11010000: Mute OFF, NORMAL Out, Stereo, Seek OFF, Bassboost
	rcall i2c_write				; R02 - High
	ldi temp1, 0b00001101		; Takt 32,768kHz, Wrap Seekdir, RDS ON, RDA5807 Enable
	rcall i2c_write				; R02 - Low
	mov temp1, Chan_Hi			; Obere 8 Bit der Startfrequenz
	rcall i2c_write				; R03 - High
	mov temp1, Chan_Lo			; Untere 2 Bit der Startfrequenz, Band 0, 100kHz Kanalraster
	rcall i2c_write				; R03 - Low
	ldi temp1, 0x08				; 50s De-emphasis wie in Europa blich
	rcall i2c_write				; R04 - High
	ldi temp1, 0x00				; Es sind nicht alle Bits beschrieben!
	rcall i2c_write				; R04 - Low
	ldi temp1, 0x88				; Interrupt = default, untere 3 Bits = Seek Threshold
	rcall i2c_write				; R05 - High
	ldi temp1, 0x80				; Bit 7 ist im Datenblatt nicht beschrieben!?
	or temp1, Volume			; untere 4 Bits fr die Lautstrke
	rcall i2c_write				; R05 - Low
	rcall i2c_stop
	pop temp1
	ret
;
fm_seek_wait:
	rcall fm_command
	ldi CtrlSeek, 0xD0			; Seek up/down Bits 0 & 1 wieder lschen
fm_wait_seek_1:
	rcall fm_read				; Doppelbyte von Register A0 einlesen
	sbrs temp2, 6				; STC Bit prfen (Seek/Tune Complete)
	rjmp fm_wait_seek_1			; warten bis STC Bit gesetzt ist
	andi temp2, 0b00000011		; nur die Bits 0 & 1 gehren zum CHANNEL.
	subi temp1, LOW(-870)		; 870 addieren (87,0MHz im 100kHz Raster)
	sbci temp2, HIGH(-870)
	sts FLo, temp1
	sts FHi, temp2
	rcall BIN2BCD_16_Out		; Frequenz ausgeben
	rcall EE_Store				; Frequenz speichern
	rcall ClearBuffer			; alten RDS Text nach Senderwechsel lschen
	ret
;
fm_read:						; Alle RDA5807 Leseregister nach SRAM auslesen
	rcall i2c_start
	ldi temp1, (0x10 << 1) | 1	; I2C Adresse fr sequenziellen Zugriff (lesend)
	rcall i2c_write
	set							; ACK anfordern => Folgebytes
;
	ldi ZL, LOW(R0AHi)			; Erstes Byte im SRAM
	ldi ZH, HIGH(R0AHi)
rlp:rcall i2c_read				; Byte nach temp1 einlesen
	st z+, temp1				; in SRAM speichern & post Inkrement Adresse
	cpi ZL, LOW(R0AHi) + 11		; vorletztes Byte erreicht ?
	brne rlp
;
	clt							; letztes Byte ohne ACK lesen
	rcall i2c_read
	st z, temp1
	rcall i2c_stop				; I2C Bus freigeben
;
	lds temp1, R0AHi			; ### Nachbearbeitung ###
	mov temp2, temp1			; Channel Bits fr fm_seek_wait sichern
	cbr Flags, 1<<Stereo		; Stereo Flag lschen
	andi temp1, 0b00000100		; Stereo Bit aus R0AHi isolieren
	or Flags, temp1				; Stereo Flag aktualisieren
;
	lds temp1, R0BHi
	mov RSSI, temp1				; obere 6 Bit = Signalstrke
	lsr RSSI
	lsr RSSI
;
	lds temp1, R0ALo			; channel Bits fr fm_seek_wait
	ret							; => temp2 = High Byte, temp1 = Low Byte
;
fm_tune:						; Sender mit Frequenz aus FLo/FHi einstellen
	lds temp1, FLo				; einzustellende Frequenz laden
	lds temp2, FHi
	subi temp1, LOW(870)		; und in RDA5807 CHANNEL Bits transformieren
	sbci temp2, HIGH(870)
	andi temp2, 0b00000011		; nur die unteren 2 Bits gehren dazu
;
	ldi Counter, 6
tune_lp:
	lsl temp1					; 6x nach links schieben
	rol temp2
	dec Counter
	brne tune_lp
	ori temp1, 0b00010000		; Tune Bit setzen
	mov Chan_Hi, temp2			; RDA5807 Register
	mov Chan_Lo, temp1
	rcall fm_command
	andi Chan_Lo, 0b11101111	; Tune Bit lschen
	ret
;
;**********************************************************
;* Basisfunktionen zur Ansteuerung des I2C Bus als Master *
;**********************************************************
i2c_init:						; Portbits setzen
	cbi PORTB, 0
	cbi PORTB, 1
	rcall scl_hi
	rcall sda_hi
	ret
;
i2c_start:
	rcall sda_lo
	rcall scl_lo
	ret
;
i2c_stop:
	rcall scl_hi
	rcall sda_hi
	ret
;
i2c_write:						; Byte aus temp1 auf I2C ausgeben
	push temp1
	push temp2
	ldi temp2, 8				; 8 Bit seriell schreiben
i2c_wr_0:
	sbrc temp1, 7				; oberes Bit zuerst
	rcall sda_hi
	sbrs temp1, 7
	rcall sda_lo
	rcall scl_hi
	rcall scl_lo
	lsl temp1
	dec temp2
	brne i2c_wr_0
	rcall sda_hi
	rcall scl_hi
	rcall scl_lo
	pop temp2
	pop temp1
	ret
;
i2c_read:						; Byte von I2C nach temp1 einlesen
	push temp2
	ldi temp2, 8				; 8 Bit seriell lesen
	rcall sda_hi
i2c_rd_0:
	rcall scl_hi
	lsl temp1					; schieben
	sbic PINB, 0				; SDA abfragen
	inc temp1					; wenn Bit=1: speichern
	rcall scl_lo
	dec temp2
	brne i2c_rd_0
	brtc i2c_read_skip_ack		; je nach Transfer Flag Ack senden
	rcall sda_lo
i2c_read_skip_ack:
	rcall scl_hi
	rcall scl_lo
	rcall sda_lo
	pop temp2
	ret
;
i2c_delay:
	push temp1
	ldi temp1, 7
i2c_delay_0:
	dec temp1
	brne i2c_delay_0
	pop temp1
	ret
;
scl_lo:
	sbi DDRB, 1
	rcall i2c_delay
	ret
;
scl_hi:
	cbi DDRB, 1
	rcall i2c_delay
	ret
;
sda_lo:
	sbi DDRB, 0
	rcall i2c_delay
	ret
;
sda_hi:
	cbi DDRB, 0
	rcall i2c_delay
	ret
;
;**************************************
;* OLED Routinen (SH1106 und SSD1306) *
;**************************************
OLED_Start_CMD:					; OLED Commando Modus per I2C starten
	rcall i2c_start
	ldi temp1, 0x78				; Basisadresse zum Schreiben
	rcall i2c_write
	ldi temp1, 00				; 0 = OLED CMD Modus
	rcall i2c_write
	ret
;
OLED_Start_Data:				; OLED Daten Modus per I2C starten
	rcall i2c_start
	ldi temp1, 0x78				; Basisadresse zum Schreiben
	rcall i2c_write
	ldi temp1, 0x40				; 0x40 = OLED Data Modus
	rcall i2c_write
	ret
;
OLED_init:
	rcall OLED_Start_CMD		; OLED Commando Modus per I2C starten
	ldi ZL, low (IniTab*2)		; Tabelle mit INIT Bytefolge
	clr ZH						; Tabellen liegen unterhalb von Adresse 256
	ldi Counter, 5				; Tabelle hat 5 Byte + 1 Fllbyte am Ende
InitLp:
	lpm temp1, Z+
	rcall i2c_write				; an OLED ausgeben
	dec Counter
	brne InitLp
	rcall i2c_stop
;	
	clr oled_y					; Display lschen
	ldi temp3, 8				; Zeilenzhler
ClrDisplay:
	clr oled_x
	rcall OLED_GotoXY
	ldi ZL, LOW(Blank*2)		; 16 Leerzeichen
	rcall OLED_PutString
	inc oled_y
	dec temp3
	brne ClrDisplay
	ret
;
OLED_Home:						; Cursor Home (oben links)
	ldi oled_x, 0
	ldi oled_y, 0
	rcall OLED_GotoXY
	ret
;
OLED_GotoXY:					; Cursor auf oled_x, oled_y setzen
	rcall OLED_Start_CMD		; OLED Commando Modus per I2C starten
	mov temp2, oled_y
	ldi temp1, 7
	sub temp1, temp2			; temp1 = 7 - oled_y
	andi temp1, 0x0F			; Maske lower Nibble
	ori temp1, 0xB0				; Page Start Konstante
	rcall i2c_write
;
	mov temp1, oled_x
	lsl temp1
	lsl temp1
	lsl temp1					; temp1 = 8 * oled_x
;	subi temp1, -2				; 2 Spalten addieren (fr SH1106 OLED)
	mov temp2, temp1			; Wert fr 2. Durchgang retten
	swap temp1					; oberes Nibble nach unten
	andi temp1, 0x0F			; Rest lschen
	ori temp1, 0x10				; Higher Nibble von Col. Start Addr
	rcall i2c_write
	mov temp1, temp2			; Wert von oben zurck
	andi temp1, 0x0F			; Lower Nibble von Col. Start Addr.
	rcall i2c_write
	rcall i2c_stop
	ret
;
Point_to_Char:					; Z auf Beginn des auszugebenden Zeichens setzen
	ldi zl, low(Font*2)			; Zeiger auf Beginn der Font Tabelle
	ldi zh, high(Font*2)
	subi temp1, 32				; Die Font Tabelle beginnt mit ' ' = Chr(32)
	clr yh						; Offset zum Zeichenbeginn in Font Tabelle berechnen
	mov yl, temp1
	ldi counter, 3				; 2^3 = 8
LP1:lsl yl						; C		<= B7..B0 <= 0
	rol yh						; Cneu  <= B7..B0 <= Calt
	dec counter
	brne LP1
	add zl, yl
	adc zh, yh					; Z steht jetzt auf Zeichenbeginn in Font Tabelle
	ldi counter, 8				; jedes Zeichen hat 8 Byte
	ret
;
OLED_PutChar:					; Zeichen aus temp1 ab Cursorposition auf OLED schreiben
	rcall Point_to_char
	rcall OLED_Start_Data
LP2:lpm temp1, z+				; Byte aus Tabelle holen und Zeiger inkrementieren
	rcall i2c_write				; Byte an Display senden
	dec counter
	brne LP2
	rcall i2c_stop
	ret
;
OLED_PutString:					; String ab Pointer Z bis zur 0 ausgeben
	lpm temp1, z+				; Buchstabe laden und post-increment Adresse
	cpi temp1, 0
	breq PutStringExit
	push zl						; LPM funktioniert nur mit den Z-Registern
	push zh						; Z sichern, da es in OLED_PutChar auch benutzt wird
	rcall OLED_PutChar
	pop zh						; altes Z von hier zurck
	pop zl
	rjmp OLED_PutString
PutStringExit:
	ret
;
;	Doppelt groe Zeichen ausgeben
stretch:						; 1 Nibble aus Zeichentabelle in temp1 auf 1 Byte dehnen
	bst	temp1, 0				; Bits ber Transfer Flag umkopieren
	bld    r2, 0				; Bit 0 von temp1 => Bits 0 & 1 von r2
	bld    r2, 1
	bst	temp1, 1
	bld    r2, 2
	bld    r2, 3
	bst	temp1, 2
	bld    r2, 4
	bld    r2, 5
	bst	temp1, 3
	bld    r2, 6
	bld    r2, 7
	mov temp1, r2				; zurck nach temp1
	ret
;
OLED_PutBigChar:				; Zeichen aus temp1 ab Cursorposition auf OLED schreiben
	rcall Point_to_char
	push zl
	push zh						; Zeiger fr 2. Durchgang retten
	rcall OLED_Start_Data
LP4:lpm temp1, z+				; Byte aus Tabelle holen und Zeiger inkrementieren
	swap temp1
	rcall stretch				; oberes Nibble auf 1 Byte dehnen
	push temp1					; Byte retten (wird in i2c_write zerstrt)
	rcall i2c_write				; Byte an Display senden
	pop temp1					; Byte zurck
	rcall i2c_write				; nochmal ausgeben
	dec counter
	brne LP4
	rcall i2c_stop
;
;	2. Durchgang
	inc oled_y					; 1 Zeile tiefer
	rcall OLED_GotoXY
	pop zh
	pop zl						; Zeiger auf Zeichen zurck
	ldi counter, 8
	rcall OLED_Start_Data
ZL3:lpm temp1, z+
	rcall stretch				; unteres Nibble auf 1 Byte dehnen
	push temp1
	rcall i2c_write
	pop temp1
	rcall i2c_write
	dec counter
	brne ZL3
	rcall i2c_stop
;
	dec oled_y					; Cursor Position aktualisieren fr nchstes Zeichen
	subi oled_x, -2				; 2 addieren
	rcall OLED_GotoXY
	ret
;
;**********************************************************************************
;* 16 Bit Binr zu BCD Wandler: Input: temp1, temp2, Output: BCD-Code in D01..D05 *
;* Danach Ausgabe von D01 bis D05 auf OLED oben links mit Leading-Zero Blanking   *
;**********************************************************************************
BIN2BCD_16_Out:
	ldi	ZL,LOW(D01)				; Zeiger auf Einerstelle im SRAM
	ldi ZH,HIGH(D01)
	ldi	counter, 0
dig:sub	counter,counter			; Divisionsrest & Carry lschen
	ldi	temp3, 16 + 1			; Schleifenzhler= Zahl der Bits + 1
dig1:
	rol	temp1					; temp1, temp2 (16 Bit) x 2
	rol	temp2
	dec	temp3					; decrement Bitzhler
	breq dig3					; Digit speichern, wenn alle Bits verarbeitet wurden
	rol	counter					; shift dividend into remainder
	subi counter, 10			; remainder = remainder - divisor
	brcc dig2					; falls < 0
	subi counter, -10			; Remainder zurck holen
	clc							; clear carry to be shifted into result
	rjmp dig1					; else
dig2:
	sec							; set carry to be shifted into result
	rjmp dig1
dig3:
	subi counter, -48			; ASCII Korrektur fr OLED Ausgabe
	st	Z+, counter				; mit post Inkrement im Puffer speichern
	cpi	ZL, LOW(D01) +5			; Digit5 erreicht ? (ZH bleibt unverndert)
	brne dig
;
;	Frequenz ausgeben
	rcall OLED_Home				; Cursor oben links
	lds temp1, D04				; '0' = 110000, '1' = 110001
	sbrc temp1, 0				; fhrende Null unterdrcken
	ldi temp1, '1'				; Wert kann nur 0 oder 1 sein!
	sbrs temp1, 0
	ldi temp1, ' '
	rcall OLED_PutBigChar
	lds temp1, D03
	rcall OLED_PutBigChar
	lds temp1, D02
	rcall OLED_PutBigChar
	ldi temp1, '.'
	rcall OLED_PutBigChar
	lds temp1, D01
	rcall OLED_PutBigChar
	ret
;
;		### EEPROM lesen und schreiben (hierzu ggf. Interrupts sperren) ###
;
EE_Read:						; 1 Byte aus Adresse temp2 nach temp1 einlesen
	sbic EECR, EEPE				; prfe ob vorheriger Schreibzugriff beendet ist
	rjmp EE_Read				; wenn nicht, weiter prfen
;	out EEARH, NULL				; Adresse in EEPROM Adressregister laden, High hier immer = 0
	out EEARL, temp2			; Low Adressbyte gem Vorgabe
	sbi EECR, EERE				; Lesen aktivieren EEPROM-Read-Bit im EEPROM Controlreg.
	in temp1, EEDR				; Daten aus EEPROM Dataregister in C Register kopieren
	ret

EE_Write:						; 1 Byte aus temp1 nach Adresse temp2 schreiben
	sbic EECR, EEPE				; prfe ob der letzte Schreibzugriff beendet ist
	rjmp EE_Write				; wenn nicht, weiter prfen
	ldi temp3, (0<<EEPM1)|(0<<EEPM0)	; Modus: Lschen und (ber)schreiben
	out EECR, temp3
;	out EEARH, NULL				; Adresse in EEPROM Adressregister laden, High hier immer = 0
	out EEARL, temp2			; Low Adressbyte gem Vorgabe
	out EEDR, temp1				; Datenbyte in EEPROM Datenregister schreiben
	sbi EECR, EEMPE				; Schreiben vorbereiten
	sbi EECR, EEPE				; Und los !
	ret
;
EE_Store:						; Empfangsfrequenz (2 Byte) ins EEPROM speichern
	ldi temp2, Last_FM + 4		; Speicheradresse im EEPROM
	lds temp1, FLO				; Low Byte der aktuellen Empfangsfrequenz
	rcall EE_Write				; ins EEPROM schreiben
	inc temp2					; dto nchstes Byte fr High Byte der Frequenz
	lds temp1, FHI
	rcall EE_Write
	ret
;
;	Font Tabelle: Zeichen 32 (' ') - 122 ('z')
.include "font_reduced.asm"
;
.eseg							; Daten im EEPROM definieren
Last_FM:	.db 0, 0, 0, 0, 0, 0; Bytes 0-3 leer, Byte 4 = FLo, Byte 5 = FHi
